import atexit
from optparse import OptionParser
import os
from subprocess import PIPE
from subprocess import Popen
import sys

from instrumental.importer import ImportHook
from instrumental.instrument import AnnotatorFactory
from instrumental.monkey import monkey_patch_imp
from instrumental.recorder import ExecutionRecorder
from instrumental.reporting import ExecutionReport

parser = OptionParser(usage="instrumental [options] COMMAND ARG1 ARG2 ...")
parser.disable_interspersed_args()
parser.add_option('-r', '--report', dest='report',
                  action='store_true',
                  help='Print a detailed coverage report')
parser.add_option('-s', '--summary', dest='summary',
                  action='store_true',
                  help='Print a summary coverage report')
parser.add_option('-S', '--statements', dest='statements',
                  action='store_true',
                  help='Print a summary statement coverage report')
parser.add_option('-x', '--xml', dest='xml',
                  action='store_true',
                  help='Create a cobertura-compatible xml coverage report')
parser.add_option('--html', dest='html',
                  action='store_true',
                  help='Create an html coverage report')
parser.add_option('-a', '--all', dest='all',
                  action='store_true', default=False,
                  help='Show all constructs (not just those missing coverage')
parser.add_option('-t', '--target', dest='targets',
                  action='append', default=[],
                  help=('A Python regular expression; modules with names'
                        ' matching this regular expression will be'
                        ' instrumented and have their coverage reported'))
parser.add_option('-i', '--ignore', dest='ignores',
                  action='append', default=[],
                  help=('A Python regular expression; modules with names'
                        ' matching this regular expression will be'
                        ' ignored and not have their coverage reported'))

def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    
    opts, args = parser.parse_args(argv)
    
    if len(args) < 1:
        parser.print_help()
        sys.exit()
    
    if not opts.targets:
        print "No targets specified. Use the '-t' option to specify packages to cover"
        sys.exit()
    
    recorder = ExecutionRecorder.get()
    annotator_factory = AnnotatorFactory(recorder)
    monkey_patch_imp(opts.targets, opts.ignores, annotator_factory)
    for target in opts.targets:
        sys.meta_path.append(ImportHook(target, opts.ignores, annotator_factory))
    
    xml_filename = os.path.abspath('instrumental.xml')
    
    sourcefile = args[0]
    environment = {'__name__': '__main__',
                   '__file__': sourcefile,
                   }
    sys.argv = args[:]
    try:
        here = os.getcwd()
        execfile(sourcefile, environment)
    finally:
        if any([opts.summary,
                opts.report,
                opts.statements,
                opts.xml,
                opts.html]):
            print
            report = ExecutionReport(here, recorder.constructs, recorder.statements, recorder.sources)
            if opts.summary:
                print report.summary()
            if opts.report:
                print report.report(opts.all)
            if opts.statements:
                print report.statement_summary()
            if opts.xml:
                report.write_xml_coverage_report(xml_filename)
            if opts.html:
                report.write_html_coverage_report()
            print



import imp
import os
import re
import sys

from instrumental.compat import ast

_imp_load_module = imp.load_module
def monkey_patch_imp(targets, ignores, visitor_factory):
    imp.load_module = load_module_factory(targets, ignores, visitor_factory)

def load_module_factory(targets, ignores, visitor_factory):
    def load_module(name, fh, pathname, description):
        if ((not any([re.match(target, name) for target in targets]))
            or
            (any([re.match(ignore, name) for ignore in ignores]))):
            return _imp_load_module(name, fh, pathname, description)
        else:
            suffix, mode, type = description
            ispkg = type == imp.PKG_DIRECTORY
            if ispkg:
                source = file(os.path.join(pathname, '__init__.py'), 'r').read()
            else:
                source = fh.read()
            visitor_factory.recorder.add_source(name, source)
            code_tree = ast.parse(source)
            visitor = visitor_factory.create(name)
            new_code_tree = visitor.visit(code_tree)
            code = compile(new_code_tree, pathname, 'exec')
            mod = sys.modules.setdefault(name, imp.new_module(name))
            if ispkg:
                mod.__file__ = os.path.join(pathname, '__init__.py')
                mod.__path__ = [pathname]
            else:
                mod.__file__ = pathname
            mod.__loader__ = None
            exec code in mod.__dict__
            return mod
    return load_module



import ast
import inspect

from astkit.render import SourceCodeRenderer as renderer

from instrumental.instrument import CoverageAnnotator
from instrumental.recorder import ExecutionRecorder

def load_module(func):
    source = inspect.getsource(func)
    normal_source =\
        "\n".join(line[12:] for line in source.splitlines(False)[1:])
    module = ast.parse(normal_source)
    return module, normal_source

class TestInstrumentation(object):
    
    def setup(self):
        # First clear out the recorder so that we'll create a new one
        ExecutionRecorder._instance = None
        self.recorder = ExecutionRecorder.get()
    
    def _load_and_compile_module(self, module_func):
        module, source = load_module(module_func)
        self.recorder.add_source(module_func.__name__, source)
        transformer = CoverageAnnotator(module_func.__name__,
                                        self.recorder)
        inst_module = transformer.visit(module)
        print renderer.render(inst_module)
        #for node in ast.walk(inst_module):
        #    print node, node.__dict__
        code = compile(inst_module, '<string>', 'exec')
        return code
    
    def test_two_pin_and_t_t(self):
        def test_module():
            a = True
            b = True
            result = a and b
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert True == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert recorder._constructs[label].conditions[0]
        assert not recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
    
    def test_two_pin_and_t_f(self):
        def test_module():
            a = True
            b = False
            result = a and b
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert False == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[0]
        assert not recorder._constructs[label].conditions[1]
        assert recorder._constructs[label].conditions[2]
    
    def test_two_pin_and_f_t(self):
        def test_module():
            a = False
            b = True
            result = a and b
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert False == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[0]
        assert recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
    
    def test_two_pin_and_f_f(self):
        def test_module():
            a = False
            b = False
            result = a and b
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert False == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[0]
        assert recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
    
    def test_two_pin_or_t_t(self):
        def test_module():
            a = True
            b = True
            result = a or b
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert True == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        print recorder._constructs[label].pins
        assert recorder._constructs[label].conditions[0]
        assert not recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
    
    def test_two_pin_or_t_f(self):
        def test_module():
            a = True
            b = False
            result = a or b
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert True == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        print recorder._constructs[label].pins
        assert recorder._constructs[label].conditions[0]
        assert not recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
    
    def test_two_pin_or_f_t(self):
        def test_module():
            a = False
            b = True
            result = a or b
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert True == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        print recorder._constructs[label].pins
        assert not recorder._constructs[label].conditions[0]
        assert recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
    
    def test_two_pin_or_f_f(self):
        def test_module():
            a = False
            b = False
            result = a or b
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert False == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        print recorder._constructs[label].pins
        assert not recorder._constructs[label].conditions[0]
        assert not recorder._constructs[label].conditions[1]
        assert recorder._constructs[label].conditions[2]
    
    def test_three_pin_and_t_t_t(self):
        def test_module():
            a = True
            b = True
            c = True
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert True == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert recorder._constructs[label].conditions[0]
        assert not recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
        assert not recorder._constructs[label].conditions[3]
    
    def test_three_pin_and_f_t_t(self):
        def test_module():
            a = False
            b = True
            c = True
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert False == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[0]
        assert recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
        assert not recorder._constructs[label].conditions[3]
    
    def test_three_pin_and_t_f_t(self):
        def test_module():
            a = True
            b = False
            c = True
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert False == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[0]
        assert not recorder._constructs[label].conditions[1]
        assert recorder._constructs[label].conditions[2]
        assert not recorder._constructs[label].conditions[3]
    
    def test_three_pin_and_t_t_f(self):
        def test_module():
            a = True
            b = True
            c = False
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert False == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[0]
        assert not recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
        assert recorder._constructs[label].conditions[3]
    
    def test_three_pin_and_f_f_f(self):
        def test_module():
            a = False
            b = False
            c = False
            result = a and b and c
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert False == result
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[0]
        assert recorder._constructs[label].conditions[1]
        assert not recorder._constructs[label].conditions[2]
        assert not recorder._constructs[label].conditions[3]
    
    def test_instrument_if_true_result(self):
        def test_module():
            a = True
            if a:
                result = 1
            else:
                result = 7
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert result is 1
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert recorder._constructs[label].conditions[True]
        assert not recorder._constructs[label].conditions[False]
        
    def test_instrument_if_false_result(self):
        def test_module():
            a = False
            if a:
                result = 1
            else:
                result = 7
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert result is 7
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[True]
        assert recorder._constructs[label].conditions[False]
        
    def test_instrument_while_false_result(self):
        def test_module():
            a = 0
            result = 0
            while a:
                a -= 1
                result += 1
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert result is 0
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert not recorder._constructs[label].conditions[True]
        assert recorder._constructs[label].conditions[False]
        
    def test_instrument_while_with_loop_result(self):
        def test_module():
            a = 3
            result = 0
            while a:
                a -= 1
                result += 1
        code = self._load_and_compile_module(test_module)
        exec code in globals(), locals()
        assert result is 3
        
        recorder = self.recorder
        label = 1
        assert label in recorder._constructs
        assert recorder._constructs[label].conditions[True]
        assert recorder._constructs[label].conditions[False]
    



import imp
import os
import re
import sys

from astkit.render import SourceCodeRenderer

from instrumental.compat import ast

class ModuleLoader(object):
    
    def __init__(self, fullpath, visitor_factory):
        self.fullpath = fullpath
        self.visitor_factory = visitor_factory
    
    def _get_source(self, path):
        return file(path, 'r').read()
    
    def _get_code(self, fullname):
        ispkg = self.fullpath.endswith('__init__.py')
        code_str = self._get_source(self.fullpath)
        self.visitor_factory.recorder.add_source(fullname, code_str)
        code_tree = ast.parse(code_str)
        visitor = self.visitor_factory.create(fullname)
        new_code_tree = visitor.visit(code_tree)
        # print SourceCodeRenderer.render(new_code_tree)
        code = compile(new_code_tree, self.fullpath, 'exec')
        return ispkg, code
    
    def load_module(self, fullname):
        ispkg, code = self._get_code(fullname)
        mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
        mod.__file__ = self.fullpath
        mod.__loader__ = self
        if ispkg:
            mod.__path__ = [os.path.dirname(self.fullpath)]
        exec code in mod.__dict__
        return mod

class ImportHook(object):
    
    def __init__(self, target, ignores, visitor_factory):
        self.target = target
        self.ignores = ignores
        self.visitor_factory = visitor_factory
    
    def find_module(self, fullname, path=[]):
        # print "find_module(%s, path=%r)" % (fullname, path), 'pyramid' in sys.modules
        if ((not re.match(self.target, fullname)) 
            or
            any([re.match(ignore, fullname) for ignore in self.ignores])):
            return None
        
        if not path:
            path = sys.path
        
        for directory in path:
            loader = self._loader_for_path(directory, fullname)
            if loader:
                return loader
    
    def _loader_for_path(self, directory, fullname):
        # print "loader_for_path", directory, fullname
        module_path = os.path.join(directory, fullname.split('.')[-1]) + ".py"
        if os.path.exists(module_path):
            # print "loading module", module_path
            loader = ModuleLoader(module_path, self.visitor_factory)
            return loader
        
        package_path = os.path.join(directory, fullname.split('.')[-1], '__init__.py')
        if os.path.exists(package_path):
            print "loading package", package_path
            loader = ModuleLoader(package_path, self.visitor_factory)
            return loader
        



import inspect

from astkit.render import SourceCodeRenderer as renderer

from instrumental.compat import ast
from instrumental.instrument import CoverageAnnotator
from instrumental.recorder import ExecutionRecorder

def load_module(func):
    source = inspect.getsource(func)
    normal_source =\
        "\n".join(line[12:] for line in source.splitlines(False)[1:])
    module = ast.parse(normal_source)
    return module, normal_source

class TestInstrumentNodesPython2(object):
    
    def setup(self):
        # First clear out the recorder so that we'll create a new one
        ExecutionRecorder._instance = None
        self.recorder = ExecutionRecorder.get()
    
    def _instrument_module(self, module_func):
        module, source = load_module(module_func)
        self.recorder.add_source(module_func.__name__, source)
        transformer = CoverageAnnotator(module_func.__name__,
                                        self.recorder)
        inst_module = transformer.visit(module)
        print renderer.render(inst_module)
        return inst_module
    
    def _assert_recorder_setup(self, module):
        assert isinstance(module, ast.Module)
        
        assert isinstance(module.body[0], ast.ImportFrom)
        assert module.body[0].module == 'instrumental.recorder'
        assert isinstance(module.body[0].names[0], ast.alias)
        assert module.body[0].names[0].name == 'ExecutionRecorder'
        
        assert isinstance(module.body[1], ast.Assign)
        assert isinstance(module.body[1].targets[0], ast.Name)
        assert module.body[1].targets[0].id == '_xxx_recorder_xxx_'
        assert isinstance(module.body[1].value, ast.Call)
        assert isinstance(module.body[1].value.func, ast.Attribute)
        assert isinstance(module.body[1].value.func.value, ast.Name)
        assert module.body[1].value.func.value.id == 'ExecutionRecorder'
        assert module.body[1].value.func.attr == 'get'
        assert not module.body[1].value.args
        assert not module.body[1].value.keywords
        assert not module.body[1].value.starargs
        assert not module.body[1].value.kwargs
    
    def _assert_record_statement(self, statement, modname, lineno):
        assert isinstance(statement, ast.Expr)
        assert isinstance(statement.value, ast.Call)
        assert isinstance(statement.value.func, ast.Attribute)
        assert isinstance(statement.value.func.value, ast.Name)
        assert statement.value.func.value.id == '_xxx_recorder_xxx_'
        assert statement.value.func.attr == 'record_statement'
        assert isinstance(statement.value.args[0], ast.Str)
        assert statement.value.args[0].s == modname
        assert isinstance(statement.value.args[1], ast.Num)
        assert statement.value.args[1].n == lineno
    
    def test_simple_module(self):
        def test_module():
            a = True
            b = True
            result = a and b
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
    def test_FunctionDef(self):
        def test_module():
            def foo():
                bar = 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.FunctionDef)
        assert inst_module.body[3].name == 'foo'
        assert isinstance(inst_module.body[3].args, ast.arguments)
        assert not inst_module.body[3].args.args
        assert not inst_module.body[3].args.vararg
        assert not inst_module.body[3].args.kwarg
        assert not inst_module.body[3].args.defaults
        assert 2 == len(inst_module.body[3].body)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Assign)
        assert isinstance(inst_module.body[3].body[1].targets[0], ast.Name)
        assert inst_module.body[3].body[1].targets[0].id == 'bar'
        assert isinstance(inst_module.body[3].body[1].value, ast.Num)
        assert inst_module.body[3].body[1].value.n == 4
    
    def test_ClassDef(self):
        def test_module():
            class FooClass(object):
                bar = 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.ClassDef)
        assert inst_module.body[3].name == 'FooClass'
        assert isinstance(inst_module.body[3].bases[0], ast.Name)
        assert inst_module.body[3].bases[0].id == 'object'
        assert 2 == len(inst_module.body[3].body)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Assign)
        assert isinstance(inst_module.body[3].body[1].targets[0], ast.Name)
        assert inst_module.body[3].body[1].targets[0].id == 'bar'
        assert isinstance(inst_module.body[3].body[1].value, ast.Num)
        assert inst_module.body[3].body[1].value.n == 4
    
    def test_Return(self):
        def test_module():
            def foo():
                return 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.FunctionDef)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Return)
        assert isinstance(inst_module.body[3].body[1].value, ast.Num)
        assert inst_module.body[3].body[1].value.n == 4
    
    def test_Delete(self):
        def test_module():
            def foo():
                del bar
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.FunctionDef)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1], ast.Delete)
        assert isinstance(inst_module.body[3].body[1].targets[0], ast.Name)
        assert inst_module.body[3].body[1].targets[0].id == 'bar'
    
    def test_Assign(self):
        def test_module():
            a = True
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.Assign)
        assert isinstance(inst_module.body[3].targets[0], ast.Name)
        assert inst_module.body[3].targets[0].id == 'a'
        assert isinstance(inst_module.body[3].value, ast.Name)
        assert inst_module.body[3].value.id == 'True'
    
    def test_AugAssign(self):
        def test_module():
            a += 4
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.AugAssign)
        assert isinstance(inst_module.body[3].target, ast.Name)
        assert inst_module.body[3].target.id == 'a'
        assert isinstance(inst_module.body[3].op, ast.Add)
        assert isinstance(inst_module.body[3].value, ast.Num)
        assert inst_module.body[3].value.n == 4
    
    def test_Print(self):
        def test_module():
            print bar
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.Print)
        assert not inst_module.body[3].dest
        assert isinstance(inst_module.body[3].values[0], ast.Name)
        assert inst_module.body[3].values[0].id == 'bar'
        assert inst_module.body[3].nl
    
    def test_For(self):
        def test_module():
            for i in [1,2,3,5]:
                print i
            else:
                print 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.For)
        assert isinstance(inst_module.body[3].target, ast.Name)
        assert inst_module.body[3].target.id == 'i'
        assert isinstance(inst_module.body[3].iter, ast.List)
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1],ast.Print)
        self._assert_record_statement(inst_module.body[3].orelse[0], 'test_module', 4)
        assert isinstance(inst_module.body[3].orelse[1],ast.Print)
    
    def test_While(self):
        def test_module():
            while i:
                print i
            else:
                print 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.While)
        assert isinstance(inst_module.body[3].test, ast.Call)
        assert isinstance(inst_module.body[3].test.func, ast.Attribute)
        assert isinstance(inst_module.body[3].test.func.value, ast.Name)
        assert inst_module.body[3].test.func.value.id == '_xxx_recorder_xxx_'
        assert inst_module.body[3].test.func.attr == 'record'
        assert isinstance(inst_module.body[3].test.args[0], ast.Name)
        assert inst_module.body[3].test.args[0].id == 'i'
        assert isinstance(inst_module.body[3].test.args[1], ast.Num)
        assert inst_module.body[3].test.args[1].n == 1
        assert not inst_module.body[3].test.keywords
        assert not hasattr(inst_module.body[3].test, 'starargs')
        assert not hasattr(inst_module.body[3].test, 'kwargs')
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1],ast.Print)
        self._assert_record_statement(inst_module.body[3].orelse[0], 'test_module', 4)
        assert isinstance(inst_module.body[3].orelse[1],ast.Print)

    def test_If(self):
        def test_module():
            if i:
                print i
            else:
                print 'else'
        inst_module = self._instrument_module(test_module)
        self._assert_recorder_setup(inst_module)
        
        self._assert_record_statement(inst_module.body[2], 'test_module', 1)
        assert isinstance(inst_module.body[3], ast.If)
        assert isinstance(inst_module.body[3].test, ast.Call)
        assert isinstance(inst_module.body[3].test.func, ast.Attribute)
        assert isinstance(inst_module.body[3].test.func.value, ast.Name)
        assert inst_module.body[3].test.func.value.id == '_xxx_recorder_xxx_'
        assert inst_module.body[3].test.func.attr == 'record'
        assert isinstance(inst_module.body[3].test.args[0], ast.Name)
        assert inst_module.body[3].test.args[0].id == 'i'
        assert isinstance(inst_module.body[3].test.args[1], ast.Num)
        assert inst_module.body[3].test.args[1].n == 1
        assert not inst_module.body[3].test.keywords
        assert not hasattr(inst_module.body[3].test, 'starargs')
        assert not hasattr(inst_module.body[3].test, 'kwargs')
        self._assert_record_statement(inst_module.body[3].body[0], 'test_module', 2)
        assert isinstance(inst_module.body[3].body[1],ast.Print)
        self._assert_record_statement(inst_module.body[3].orelse[0], 'test_module', 4)
        assert isinstance(inst_module.body[3].orelse[1],ast.Print)



import os
import sys
import time

from xml.etree import ElementTree

import instrumental
from instrumental.constructs import BooleanDecision

class CoverageObject(object):
    
    @property
    def line_rate(self):
        statements = self.statements
        if statements:
            total_statements = len(statements)
            hit_statements = len([statement for statement in statements
                                  if statement])
            return hit_statements / total_statements
        else:
            return 1.0
    
    @property
    def branch_rate(self):
        decisions = self.decisions
        if decisions:
            total_conditions = 2 * len(decisions)
            total_conditions_hit = sum(construct.number_of_conditions_hit()
                                       for construct in decisions)
            return total_conditions_hit / total_conditions
        else:
            return 1.0
    
    def to_element(self):
        element = ElementTree.Element(self.tag)
        if hasattr(self, 'name'):
            element.attrib['name']  = self.name
        element.attrib['complexity'] = '0.0'
        element.attrib['line-rate'] = "%f" % self.line_rate
        element.attrib['branch-rate'] = "%f" % self.branch_rate
        return element

class Coverage(CoverageObject):
    tag = 'coverage'
    
    def __init__(self, report):
        self._report = report
    
    @property
    def statements(self):
        _statements = []
        for modulename in self._report.statements:
            _statements += self._report.statements[modulename].values()
        return _statements
    
    @property
    def decisions(self):
        return [construct for construct in self._report.constructs.values()
                if isinstance(construct, BooleanDecision)]
        
    def to_element(self):
        coverage_element = super(Coverage, self).to_element()
        del coverage_element.attrib['complexity']
        coverage_element.attrib['timestamp'] = str(int(time.time()))
        coverage_element.version = instrumental.__version__
        
        comment = 'Generated by instrumental: http://bitbucket.org/desmaj/instrumental'
        coverage_element.append(ElementTree.Comment(comment))
        
        here = self._report.working_directory
        packages = {}
        for modulename in self._report.statements.keys():
            module = sys.modules[modulename]
            modules = packages.setdefault(module.__package__, [])
            modules.append((modulename, os.path.relpath(module.__file__, here)))
        
        packages_element = ElementTree.Element('packages')
        for packagename, modulenames in packages.items():
            package_coverage = PackageCoverage(packagename,
                                               packages[packagename],
                                               self._report.statements,
                                               self._report.constructs)
            package_element = package_coverage.to_element()
            packages_element.append(package_element)
        
        coverage_element.append(packages_element)
        
        return coverage_element

class PackageCoverage(CoverageObject):
    tag = 'package'
    
    def __init__(self, name, modules, statements, constructs):
        self.name = name
        self._modules = modules
        self._statements = statements
        self._constructs = constructs
    
    @property
    def statements(self):
        _statements = []
        for modulename, modulefile in self._modules:
            _statements += self._statements[modulename].values()
        return _statements
    
    @property
    def decisions(self):
        _decisions = []
        for modulename, modulefile in self._modules:
            _decisions += [construct 
                           for construct in self._constructs.values()
                           if (isinstance(construct, BooleanDecision)
                               and construct.modulename==modulename)]
        return _decisions
    
    def to_element(self):
        element = super(PackageCoverage, self).to_element()
        
        modules = ElementTree.Element('classes')
        element.append(modules)
        
        for modulename, modulefile in self._modules:
            module_constructs = [construct for construct in self._constructs.values()
                                 if construct.modulename==modulename]
            module_coverage = ModuleCoverage(modulename,
                                             modulefile,
                                             self._statements[modulename],
                                             module_constructs)
            modules.append(module_coverage.to_element())
        
        return element
        
class ModuleCoverage(CoverageObject):
    tag = 'class'
    
    def __init__(self, name, filename, statements, constructs):
        self.name = name
        self.filename = filename
        self._statements = statements
        self._constructs = constructs
    
    @property
    def statements(self):
        return self._statements.values()
    
    @property
    def decisions(self):
        _decisions = []
        for construct in self._constructs:
            if isinstance(construct, BooleanDecision):
                _decisions.append(construct)
        return _decisions
    
    def to_element(self):
        element = super(ModuleCoverage, self).to_element()
        element.attrib['filename'] = self.filename
        element.append(ElementTree.Element('methods'))
        
        lines_element = ElementTree.Element('lines')
        for lineno, hit in sorted(self._statements.items()):
            line_element = ElementTree.Element('line')
            line_element.attrib['line'] = str(lineno)
            line_element.attrib['hits'] = str(int(hit))
            lines_element.append(line_element)
        element.append(lines_element)
        
        return element
    
class XMLCoverageReport(object):
    
    def __init__(self, report):
        self._report = report
    
    def write(self, filename):
        tree = str(self)
        document = '\n'.join([
                '<?xml version="1.0" ?>',
                '<!DOCTYPE coverage ',
                "  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>",
                tree,
                ])
        file(filename, 'w').write(document)
    
    def __str__(self):
        return ElementTree.tostring(self.tree.getroot())
    
    @property
    def tree(self):
        coverage_element = Coverage(self._report)
        coverage = coverage_element.to_element()
        
        _tree = ElementTree.ElementTree(coverage)
        return _tree
        



from copy import copy
from copy import deepcopy
import inspect
import sys

from instrumental.compat import ast
from instrumental.constructs import BooleanDecision
from instrumental.constructs import LogicalAnd
from instrumental.constructs import LogicalOr
from instrumental.pragmas import PragmaFinder

def __setup_recorder(): # pragma: no cover
    from instrumental.recorder import ExecutionRecorder
    _xxx_recorder_xxx_ = ExecutionRecorder.get()

def get_setup():
    source = inspect.getsource(__setup_recorder)
    mod = ast.parse(source)
    defn = mod.body[0]
    setup = defn.body[:]
    for stmt in setup:
        stmt.lineno -= 1
    return setup

class ExecutionRecorder(object):
    
    _instance = None
    @classmethod
    def get(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance
    
    def __init__(self):
        self._next_label = 1
        self._constructs = {}
        self._sources = {}
        self._statements = {}
        self._branches = {}
        self.pragmas = {}
    
    def add_source(self, modulename, source):
        self._sources[modulename] = source
        self.pragmas[modulename] = self._find_pragmas(source)
    
    def _find_pragmas(self, source):
        finder = PragmaFinder()
        return finder.find_pragmas(source)
    
    @property
    def sources(self):
        return self._sources
    
    @property
    def constructs(self):
        return self._constructs
    
    @property
    def statements(self):
        return self._statements
    
    def next_label(self):
        label = self._next_label
        self._next_label += 1
        return label
    
    @staticmethod
    def get_recorder_call():
        kall = ast.Call()
        kall.func = ast.Attribute(value=ast.Name(id="_xxx_recorder_xxx_",
                                                 ctx=ast.Load()),
                                  attr="record",
                                  ctx=ast.Load())
        kall.keywords = []
        return kall
    
    def record(self, arg, label, *args, **kwargs):
        self._constructs[label].record(arg, *args, **kwargs)
        return arg
    
    def add_BoolOp(self, modulename, node):
        if isinstance(node.op, ast.And):
            construct_klass = LogicalAnd
        elif isinstance(node.op, ast.Or):
            construct_klass = LogicalOr
        else:
            raise TypeError("Expected a BoolOp node with an op field of ast.And or ast.Or")
        construct = construct_klass(modulename, node)
        
        label = self.next_label()
        self._constructs[label] = construct
        
        # Now wrap the individual values in recorder calls
        base_call = self.get_recorder_call()
        base_call.args = \
            [ast.Num(n=label, lineno=node.lineno, col_offset=node.col_offset)]
        for i, value in enumerate(node.values):
            # Try to determine if the condition is a literal
            # Maybe we can do something with this information?
            try:
                literal = ast.literal_eval(value)
                construct.literals[i] = literal
            except ValueError:
                pass
            recorder_call = deepcopy(base_call)
            recorder_call.args.insert(0, node.values[i])
            recorder_call.args.append(ast.copy_location(ast.Num(n=i), node.values[i]))
            node.values[i] = ast.copy_location(recorder_call, node.values[i])
        ast.fix_missing_locations(node)
        return node
    
    def add_test(self, modulename, node):
        label = self.next_label()
        construct = BooleanDecision(modulename, node)
        self._constructs[label] = construct
        
        base_call = ast.copy_location(self.get_recorder_call(),
                                      node)
        base_call.args = \
            [node,
             ast.Num(n=label, lineno=node.lineno, col_offset=node.col_offset)]
        ast.fix_missing_locations(base_call)
        return base_call
    
    @staticmethod
    def get_statement_recorder_call(modulename, lineno):
        kall = ast.Call()
        kall.func = ast.Attribute(value=ast.Name(id="_xxx_recorder_xxx_",
                                                 ctx=ast.Load()),
                                  attr="record_statement",
                                  ctx=ast.Load())
        kall.args = [ast.Str(s=modulename),
                     ast.Num(n=lineno),
                     ]
        kall.keywords = []
        kall_stmt = ast.Expr(value=kall)
        return kall_stmt
    
    def record_statement(self, modulename, lineno):
        self._statements[modulename][lineno] = True
    
    def add_statement(self, modulename, node):
        lines = self._statements.setdefault(modulename, {})
        lines[node.lineno] = False
        marker = self.get_statement_recorder_call(modulename, node.lineno)
        marker = ast.copy_location(marker, node)
        ast.fix_missing_locations(marker)
        return marker



from copy import deepcopy

from astkit.render import SourceCodeRenderer

class LogicalBoolean(object):
    
    def __init__(self, modulename, node):
        self.modulename = modulename
        self.node = deepcopy(node)
        self.lineno = node.lineno
        self.source = SourceCodeRenderer.render(node)
        self.pins = len(node.values)
        self.conditions =\
            dict((i, False) for i in range(self.pins + 1))
        self.literals = {}
    
    def number_of_conditions(self):
        return len(self.conditions)
    
    def number_of_conditions_hit(self):
        return len([value 
                    for value in self.conditions.values()
                    if value])
    
    def conditions_missed(self):
        return self.number_of_conditions() - self.number_of_conditions_hit()
    
    def _literal_warning(self):
        return ("** One or more conditions maybe not be reachable due to"
                " the presence of a literal in the decision")
        
    def result(self):
        lines = []
        name = "%s -> %s:%s < %s >" % (self.__class__.__name__, self.modulename, self.lineno, self.source)
        lines.append("%s" % (name,))
        if self.literals:
            lines.append("")
            lines.append(self._literal_warning())
        lines.append("")
        for condition in sorted(self.conditions):
            lines.append(self.description(condition) +\
                             " ==> " +\
                             str(self.conditions[condition]))
        return "\n".join(lines)

class LogicalAnd(LogicalBoolean):
    """ Stores the execution information for a Logical And
        
        For an and condition with n inputs, there will be
        n + 2 recordable conditions. Condition 0 indicates
        that all inputs are True. Conditions 1 though n
        indicate that the input in the numbered position
        is False and all inputs before it are True. All
        inputs after the numbered position are, in this
        case, considered to be "don't care" since they will
        never be evaluated.
    """
    
    def record(self, value, pin):
        """ Record that a value was seen for a particular pin """
        
        # If the pin is not the last pin in the decision and
        # the value seen is False, then we've found the pin
        # that has forced the decision False and we should
        # record that.
        if pin < (self.pins-1):
            if not value:
                self.conditions[pin+1] = True
        
        # If the pin is the last pin then we'll record that
        # it either allowed the decision to be True or it
        # is the pin that has forced the decision False.
        elif pin == (self.pins-1):
            if value:
                self.conditions[0] = True
            else:
                self.conditions[self.pins] = True
    
    def description(self, n):
        if n == 0:
            return " ".join("T" * self.pins)
        elif n < (self.pins + 1):
            acc = ""
            if n > 1:
                acc += " ".join("T" * (n - 1)) + " "
            acc += "F"
            if self.pins - n:
                acc += " " + " ".join("*" * (self.pins - n))
            return acc
        elif n == (self.pins + 1):
            return "Other"
        
class LogicalOr(LogicalBoolean):
    """ Stores the execution information for a Logical Or
        
        For an or condition with n inputs, there will be
        n + 2 recordable conditions. Conditions 0 though
        n indicate that the numbered position input is
        True. Condition n indicates that all inputs are
        False. Condition n + 1 is "Other".
    """
    
    def record(self, value, pin):
        """ Record that a value was seen for a particular pin """
        
        # If the pin is not the last pin in the decision
        # and the value we see is True, then we've found the
        # condition that forces the decision True in this case.
        # If the value we see is False then we'll ignore it
        # since this is not a significant case.
        if pin < (self.pins-1):
            if value:
                self.conditions[pin] = True
        
        # If this is the last pin then it either allowed the
        # decision to be False or forced the decision True.
        elif pin == (self.pins-1):
            if value:
                self.conditions[pin] = True
            else:
                self.conditions[self.pins] = True
        
    def description(self, n):
        acc = ""
        if n < self.pins:
            if n > 0:
                acc += " ".join("F" * n) + " "
            acc += "T"
            if self.pins - n - 1:
                acc += " " + " ".join("*" * (self.pins - n - 1))
            return acc
        elif n == self.pins:
            return " ".join("F" * self.pins)
        elif n == (self.pins + 1):
            return "Other"
    
class BooleanDecision(object):
    
    def __init__(self, modulename, node):
        self.modulename = modulename
        self.node = deepcopy(node)
        self.lineno = node.lineno
        self.source = SourceCodeRenderer.render(node)
        self.conditions = {True: False,
                           False: False}
        
    def record(self, expression):
        result = bool(expression)
        self.conditions[result] = True
        return result

    def number_of_conditions(self):
        return len(self.conditions)
    
    def number_of_conditions_hit(self):
        return len([value 
                    for value in self.conditions.values()
                    if value])
    
    def conditions_missed(self):
        return self.number_of_conditions() - self.number_of_conditions_hit()
    
    def result(self):
        lines = []
        name = "Decision -> %s:%s < %s >" % (self.modulename, self.lineno, self.source)
        lines.append("%s" % (name,))
        lines.append("")
        lines.append("T ==> %r" % self.conditions[True])
        lines.append("F ==> %r" % self.conditions[False])
        return "\n".join(lines)



from instrumental.reporting import ExecutionReport

class FakeConstruct(object):
    
    def __init__(self, label, *conditions_hit):
        self.label = label
        self.modulename = "somemodulename"
        self.lineno = 17
        self.conditions = {True: False,
                           False: False}
        for condition in conditions_hit:
            self.conditions[condition] = True
    
    def result(self):
        return "ConstructResult(%s)" % self.label
    
class TestExecutionReport(object):
    
    def setup(self):
        self.T = FakeConstruct("True only", True)
        self.F = FakeConstruct("False only", False)
        self.TF = FakeConstruct("Both", True, False)
        self.missed = FakeConstruct("Neither")
    
    def _makeOne(self, working_directory='.', constructs={}, statements={}, sources={}):
        return ExecutionReport(working_directory, constructs, statements, sources)
        
    def test_header(self):
        expected_header = """
===============================================
Instrumental Condition/Decision Coverage Report
===============================================
"""
        
        reporter = self._makeOne()
        assert expected_header in reporter.report(), reporter.report()
    
    def test_showall_true(self):
        reporter = self._makeOne('.', {1: self.missed})
        assert "ConstructResult(Neither)" in reporter.report(True), reporter.report(True)



import sys

from instrumental.constructs import BooleanDecision

def _package(modulename):
    return sys.modules[modulename].__package__

class BaseExecutionSummary(object):
    
    def __init__(self, conditions, statements):
        self.decisions = dict((label, condition) for label, condition in conditions.items()
                              if isinstance(condition, BooleanDecision))
        self.conditions = conditions
        self.statements = statements
        
    def condition_rate(self):
        total_conditions = sum(condition.number_of_conditions()
                               for label, condition in self.conditions.items())
        if not total_conditions:
            return 1.0
        hit_conditions = sum(condition.number_of_conditions_hit()
                             for label, condition in self.conditions.items())
        return hit_conditions / float(total_conditions)
    
    def decision_rate(self):
        total_conditions = sum(decision.number_of_conditions()
                               for label, decision in self.decisions.items())
        if not total_conditions:
            return 1.0
        hit_conditions = sum(decision.number_of_conditions_hit()
                             for label, decision in self.decisions.items())
        return hit_conditions / float(total_conditions)
    
    def statement_rate(self):
        all_statements = []
        for modulename, statement_dict in self.statements.items():
            all_statements += statement_dict.items()
        
        total_statements = len(all_statements)
        if not total_statements:
            return 1.0
        hit_statements = sum(hit for (lineno, hit) in all_statements)
        return hit_statements / float(total_statements)


class ExecutionSummary(BaseExecutionSummary):
    
    def __init__(self, conditions, statements):
        super(ExecutionSummary, self).__init__(conditions, statements)
        self._packages = None
    
    @property
    def packages(self):
        if self._packages is None:
            _statements = {}
            _conditions = {}
            for modulename in self.statements:
                _package_statements = \
                    _statements.setdefault(_package(modulename), {})
                _package_statements[modulename] = self.statements[modulename]
                _package_conditions = \
                    _conditions.setdefault(_package(modulename), {})
                _package_conditions.update(\
                    dict((label, condition) 
                         for label, condition in self.conditions.items()
                         if condition.modulename == modulename)
                    )
                    
            self._packages = \
                dict((packagename, 
                      PackageExecutionSummary(packagename,
                                              _conditions[packagename],
                                              _statements[packagename]))
                     for packagename in _statements)
        return self._packages

        
class PackageExecutionSummary(BaseExecutionSummary):
    
    def __init__(self, name, conditions, statements):
        super(PackageExecutionSummary, self).__init__(conditions, statements)
        self.name = name
        self._modules = None
    
    @property
    def modules(self):
        if self._modules is None:
            _conditions = dict((modulename,
                                dict((label, condition) 
                                     for label, condition in self.conditions.items()
                                     if modulename == condition.modulename))
                                for modulename in self.statements)
            self._modules = \
                dict((modulename,
                      ModuleExecutionSummary(modulename,
                                             _conditions[modulename],
                                             self.statements[modulename]))
                      for modulename in self.statements)
        return self._modules


class ModuleExecutionSummary(BaseExecutionSummary):
    
    def __init__(self, name, conditions, statements):
        self.name = name
        super(ModuleExecutionSummary, self).__init__(conditions, statements)
    
    def statement_rate(self):
        total_statements = len(self.statements)
        hit_statements = sum(hit for (lineno, hit) in self.statements.items())
        return hit_statements / float(total_statements)

